// This file is part of Doodah Band.
//
// Doodah Band is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//
// Doodah Band is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
// You should have received a copy of the GNU General Public License along with Doodah Band. If not, see <https://www.gnu.org/licenses/>.

// #include <arpa/inet.h>
#include <errno.h>
#include <fcntl.h>
#include <math.h>
#include <SDL2/SDL.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
// #include <sys/socket.h>
#include <sys/stat.h>
#include <unistd.h>

#define OV_EXCLUDE_STATIC_CALLBACKS
#include "vorbis/vorbisfile.h"
#include "vorbis/codec.h"
#define FIXED_POINT 16
#include "kiss_fftr.h"

// // Maximum size of all playlist file names.
// #define PLAYLIST_BUFFER_LENGTH 4096
// Maximum length of user's shader.
#define SHADER_SOURCE_LENGTH 16384

// Quit with an error message after SDL returns an error signal.
void abort_sdl(const char *const message) {
	fprintf(stderr, "ERROR: %s (%s)\n", message, SDL_GetError());
	SDL_Quit();
	abort();
}

// Include all other files.
// This are here because some of them need values from above.
#include "audio.c.inc"
#include "config.c.inc"
#include "default_shaders.c.inc"
#include "opengl.c.inc"

// Load fragment shader from `path` into `buffer` of given `capacity`.
// Returns shader handle if successful, zero otherwise.
GLuint load_shader(const char *const path, char *source_buffer, const size_t capacity) {
	FILE *file = fopen(SHADER_PATH, "rb");
	if (file == NULL) {
		fprintf(stderr, "ERROR: failed to open %s\n", path);
		return 0;
	}
	const size_t length = fread(source_buffer, 1, capacity, file);
	fclose(file);

	const GLuint vertex = compile_shader_stage(VERTEX_SHADER, sizeof(VERTEX_SHADER), GL_VERTEX_SHADER);
	if (vertex == 0) return 0;
	const GLuint fragment = compile_shader_stage(source_buffer, length, GL_FRAGMENT_SHADER);
	if (fragment == 0) return 0;
	const GLuint program = link_shader_stages(vertex, fragment);
	return program;
}

// Return true if file exists.
bool test_exists(const char *const path) {
	struct stat data;
	return stat(path, &data) == 0;
}

// Return true if file was modified.
// `last_modified` is used to compare against, and updated with latest time.
bool check_modified(const char *const path, time_t *last_modified) {
	struct stat data;
	if (stat(path, &data) != 0) return false;
	const bool result = data.st_mtime != *last_modified;
	*last_modified = data.st_mtime;
	return result;
}

/*
// Make TCP socket non-blocking.
void disable_blocking(int socket) {
	int flags = fcntl(socket, F_GETFL, 0);
	if (flags < 0) {
		fprintf(stderr, "ERROR: failed to disable blocking on socket\n");
		abort();
	}
	flags |= O_NONBLOCK;
	if (fcntl(socket, F_SETFL, flags) != 0) {
		fprintf(stderr, "ERROR: failed to disable blocking on socket\n");
		abort();
	}
}
*/

// Upload RGBA bitmap and return texture.
GLuint upload_texture(const size_t width, const size_t height, const uint8_t *const rgba) {
	GLuint texture = 0;
	glGenTextures(1, &texture);
	glBindTexture(GL_TEXTURE_2D, texture);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, rgba);
	glGenerateMipmap(GL_TEXTURE_2D);
	return texture;
}

// Return handle of new texture containing noise.
GLuint generate_noise_texture(void) {
	// Fill square RGBA with opaque white noise with independent colour channels.
	const size_t noise_resolution = 256;
	uint8_t noise_rgba[noise_resolution*noise_resolution*4];

	// Generate noise into buffer.
	uint32_t xorshift = 0x746f01ee;
	size_t index;
	for (index = 0; index < sizeof(noise_rgba); index += 4) {
		xorshift ^= xorshift << 13;
		xorshift ^= xorshift >> 17;
		xorshift ^= xorshift << 5;
		noise_rgba[index] = xorshift;
		noise_rgba[index+1] = xorshift >> 8;
		noise_rgba[index+2] = xorshift >> 16;
		noise_rgba[index+3] = 0xff;
	}

	return upload_texture(noise_resolution, noise_resolution, noise_rgba);
}

bool is_power_of_two(const size_t value) {
	return value != 0 && (value & (value-1)) == 0;
}

// Return handle of new texture loaded from a .bmp file.
GLuint load_bmp_texture(const char *const path) {
	SDL_Surface *const surface = SDL_LoadBMP(path);
	if (surface == NULL) {
		fprintf(stderr, "ERROR: failed to load image \"%s\".\n", path);
		return 0;
	}
	SDL_Surface *const converted = SDL_ConvertSurfaceFormat(surface, SDL_PIXELFORMAT_ABGR8888, 0);
	if (converted == NULL) abort();
	SDL_FreeSurface(surface);

	if (!(is_power_of_two(converted->w) && is_power_of_two(converted->h))) {
		fprintf(stderr, "ERROR: image \"%s\" has a non-power-of-two resolution.\n", path);
		SDL_FreeSurface(converted);
		return 0;
	}

	const GLuint texture = upload_texture(converted->w, converted->h, converted->pixels);
	SDL_FreeSurface(converted);
	return texture;
}

// Return handle of new texture for live-updating.
GLuint allocate_texture(void) {
	GLuint texture = 0;
	glGenTextures(1, &texture);
	glBindTexture(GL_TEXTURE_2D, texture);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

	return texture;
}

// Bind texture to given uniform name if available.
// `slot` is the texture slot (from zero).
void bind_texture(const GLuint program, const char *const uniform, const GLuint texture, const unsigned int slot) {
	const GLint location = glGetUniformLocation(program, uniform);
	if (location >= 0) {
		glActiveTexture(GL_TEXTURE0+slot);
		glBindTexture(GL_TEXTURE_2D, texture);
		glUniform1i(location, slot);
	}
}

int main(int argc, char *argv[]) {
	(void)argc;
	(void)argv;
	
/*
	// Start playlist server.
	int listen_socket = socket(AF_INET, SOCK_STREAM, 0);
	if (listen_socket < 0) {
		perror("ERROR: failed to open playlist socket:");
		abort();
	}
	int reuse = 1;
	if (setsockopt(listen_socket, SOL_SOCKET, SO_REUSEADDR, &reuse, sizeof(reuse)) != 0) {
		perror("ERROR: failed to allow socket reuse:");
	}
	struct sockaddr_in listen_address;
	listen_address.sin_family = AF_INET;
	listen_address.sin_addr.s_addr = htonl(INADDR_ANY);
	listen_address.sin_port = htons(1234);
	if (bind(listen_socket, (const struct sockaddr *)&listen_address, sizeof(listen_address)) != 0) {
		perror("ERROR: failed to bind playlist socket:");
		abort();
	}
	if (listen(listen_socket, 1) != 0) {
		perror("ERROR: failed to listen on playlist socket:");
		abort();
	}
	disable_blocking(listen_socket);
	int client_socket = -1; // none for now.
	char *playlist = malloc(PLAYLIST_BUFFER_LENGTH);
	size_t playlist_length = 0;
	if (playlist == NULL) abort();
*/

	// Setup SDL.
	if (SDL_Init(SDL_INIT_AUDIO | SDL_INIT_EVENTS | SDL_INIT_VIDEO) != 0) abort_sdl("failed to initialise SDL2");

	SDL_Window *window = SDL_CreateWindow("Doodah-band", SDL_WINDOWPOS_UNDEFINED, SDL_WINDOWPOS_UNDEFINED, 640, 480, SDL_WINDOW_OPENGL | SDL_WINDOW_RESIZABLE);
	if (window == NULL) abort_sdl("failed to create window");

	// Open audio output.
	audio_init();
	float amplitude = 0.0;
	int next_track = 1;

	// Setup OpenGL.
	if (SDL_GL_SetAttribute(SDL_GL_CONTEXT_PROFILE_MASK, SDL_GL_CONTEXT_PROFILE_ES) != 0 ||
		SDL_GL_SetAttribute(SDL_GL_CONTEXT_MAJOR_VERSION, 2) != 0 ||
		SDL_GL_SetAttribute(SDL_GL_CONTEXT_MINOR_VERSION, 0) != 0) abort_sdl("failed to configure OpenGL context");
	// SDL_GL_SetAttribute(SDL_GL_CONTEXT_FLAGS, SDL_GL_CONTEXT_DEBUG_FLAG);
	SDL_GLContext context = SDL_GL_CreateContext(window);
	if (context == NULL) abort_sdl("failed to create OpenGL context");
	load_gl_functions(&context);

	// Load quad.
	const float vertex_positions[] = {
		0.0, 0.0,
		1.0, 0.0,
		0.0, 1.0,
		1.0, 1.0,
	};
	glEnableVertexAttribArray(0);
	glVertexAttribPointer(0, 2, GL_FLOAT, GL_FALSE, 0, &vertex_positions);

	// Create textures.
	const GLuint noise_texture = generate_noise_texture();
	GLuint image_texture = 0;
	if (test_exists(TEXTURE_PATH)) image_texture = load_bmp_texture(TEXTURE_PATH);
	const GLuint fft_texture = allocate_texture();
	const GLuint smooth_fft_texture = allocate_texture();
	const GLuint framebuffer_texture = allocate_texture();

	// Allocate shader buffer.
	char *source_buffer = malloc(SHADER_SOURCE_LENGTH);
	if (source_buffer == NULL) abort();
	time_t last_modified = 0; // we assume the initial modification time will never be zero.
	GLuint program = 0;

	// Create fallback shader if shader file does not exist.
	if (!test_exists(SHADER_PATH)) {
		FILE *file = fopen(SHADER_PATH, "wb");
		if (file != NULL) {
			fwrite(FRAGMENT_SHADER, 1, sizeof(FRAGMENT_SHADER)-1, file);
			fclose(file);
		}
	}

	// Main loop.
	const unsigned int target_fps = 60;
	const uint64_t frame_ms = 1000/target_fps;
	bool paused = false;
	bool errors = false;
	uint8_t fft_average[ANALYSIS_WINDOW_SAMPLES] = {0};
	uint64_t frame_milliseconds_sum = 0;
	unsigned int frames_sum = 0;
	float time_seconds = 0.0;
	bool run = true;
	while (run) {
		const uint64_t frame_start = SDL_GetTicks64();

		SDL_Event event;
		while (SDL_PollEvent(&event)) {
			switch (event.type) {
				case SDL_KEYDOWN:
					if (event.key.keysym.sym == SDLK_ESCAPE) {
						run =  false;
					} else if (event.key.keysym.sym == SDLK_COMMA) {
						audio_stop();
						// TODO: rewinding from beyond end of final track jumps to previous track, when it feels like it should jump to start of final track.
						if (next_track <= 2) {
							next_track = 1;
						} else {
							next_track -= 2;
						}
					} else if (event.key.keysym.sym == SDLK_PERIOD) {
						audio_stop();
					} else if (event.key.keysym.sym == SDLK_LEFT) {
						audio_skip(-1.0);
					} else if (event.key.keysym.sym == SDLK_RIGHT) {
						audio_skip(1.0);
					} else if (event.key.keysym.sym == SDLK_UP) {
						audio_skip(10.0);
					} else if (event.key.keysym.sym == SDLK_DOWN) {
						audio_skip(-10.0);
					} else if (event.key.keysym.sym == SDLK_SPACE) {
						paused = !paused;
						audio_pause(paused);
					}
					break;
				case SDL_QUIT:
					run = false;
					break;
				case SDL_WINDOWEVENT:
					if (event.window.event == SDL_WINDOWEVENT_RESIZED) {
						glViewport(0, 0, event.window.data1, event.window.data2);
					}
					break;
				default:
					break;
			}
		}

/*
		// Accept audio file paths over TCP and add to playlist.
		if (client_socket < 0) {
			struct sockaddr address;
			socklen_t size = sizeof(address);
			client_socket = accept(listen_socket, &address, &size);
			if (client_socket >= 0) {
				disable_blocking(client_socket);
			} else if (errno != EAGAIN && errno != EWOULDBLOCK) {
				perror("ERROR: failed to accept playlist client:");
			}
		}
		if (client_socket >= 0 && playlist_length < PLAYLIST_BUFFER_LENGTH) {
			const ssize_t bytes = read(client_socket, &playlist[playlist_length], PLAYLIST_BUFFER_LENGTH-playlist_length);
			if (bytes < 0 && errno != EAGAIN && errno != EWOULDBLOCK) {
				perror("ERROR: failed to read data from playlist client:");
				close(client_socket);
				client_socket = -1;
			} else {
				playlist_length += bytes;
			}
		}
*/

		// Load and analyse audio files.
		if (next_track < argc) {
			if (audio_queue(argv[next_track])) {
				next_track ++;
			}
		}
		uint8_t fft_output[ANALYSIS_WINDOW_SAMPLES];
		audio_analyse(frame_ms, &amplitude, fft_output);

		// Keep average of FFT.
		size_t index;
		for (index = 0; index < ANALYSIS_WINDOW_SAMPLES; index ++) {
			fft_average[index] = (fft_average[index]>>1) +
				(fft_average[index]>>2) +
				(fft_average[index]>>3) +
				(fft_output[index]>>3);
		}

		// Convert FFT output to texture.
		glBindTexture(GL_TEXTURE_2D, fft_texture);
		glTexImage2D(GL_TEXTURE_2D, 0, GL_LUMINANCE, ANALYSIS_WINDOW_SAMPLES/2, 1, 0, GL_LUMINANCE, GL_UNSIGNED_BYTE, fft_output);

		glBindTexture(GL_TEXTURE_2D, smooth_fft_texture);
		glTexImage2D(GL_TEXTURE_2D, 0, GL_LUMINANCE, ANALYSIS_WINDOW_SAMPLES/2, 1, 0, GL_LUMINANCE, GL_UNSIGNED_BYTE, fft_average);

		// Reload shader when file is written.
		if (check_modified(SHADER_PATH, &last_modified)) {
			GLuint new_program = load_shader(SHADER_PATH, source_buffer, SHADER_SOURCE_LENGTH);
			errors = (new_program == 0);
			if (new_program != 0) {
				glUseProgram(new_program);
				program = new_program;
			}
		}

		// Setup GPU load measurement.
		//
		// According to comments on Stack Overflow...
		// https://stackoverflow.com/questions/24135853/at-what-point-does-vsync-wait-block
		// ...swapping buffers doesn't itself block.
		// Instead, it makes drawing commands that would modify the backbuffer wait
		// until the buffers are actually swapped.
		//
		// We want to measure actual time spent drawing (not the time it takes to
		// swap buffers) so we force OpenGL to wait until all commands are finished.
		// Hopefully, this will include swapping buffers, so that all drawing
		// commands after this will be free to execute immediately.
		// We can then wait for them to finish once they're all queued, and get a
		// moderately accurate measure of the execution time.
		//
		// This gives us a rough idea of the GPU load.
		// Hopefully.
		// On some systems.
		// ...I'm doing it this way because I've had no luck loading
		// GL_ARB_timer_query with GLES 2.
		glFinish();
		uint64_t render_start = SDL_GetTicks64();

		// Redraw.
		glClear(GL_COLOR_BUFFER_BIT);
		if (program != 0) {
			int width, height;
			SDL_GetWindowSize(window, &width, &height);

			GLint location = glGetUniformLocation(program, "resolution");
			if (location >= 0) glUniform2f(location, (float)width, (float)height);

			location = glGetUniformLocation(program, "time");
			if (location >= 0) glUniform1f(location, time_seconds);

			location = glGetUniformLocation(program, "amplitude");
			if (location >= 0) glUniform1f(location, amplitude);

			bind_texture(program, "fft", fft_texture, 0);
			bind_texture(program, "smooth_fft", smooth_fft_texture, 1);
			bind_texture(program, "noise", noise_texture, 2);
			if (image_texture != 0) bind_texture(program, "image", image_texture, 3);
			bind_texture(program, "frame", framebuffer_texture, 4);

			glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);

			// Store output for next frame.
			glBindTexture(GL_TEXTURE_2D, framebuffer_texture);
			glCopyTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, 0, 0, width, height, 0);
		}

		// Measure GPU load.
		glFinish();
		const uint64_t render_end = SDL_GetTicks64();
		frames_sum ++;
		frame_milliseconds_sum += (unsigned int)(render_end-render_start);
		if (frames_sum >= target_fps) {
			const float average_milliseconds = (float)frame_milliseconds_sum/(float)frames_sum;
			char buffer[64];
			if (errors) {
				sprintf(buffer, "Doodah Band (error)");
			} else if (average_milliseconds > 1000.0) {
				// Avoid overflow, because snprintf isn't portable.
				sprintf(buffer, "Doodah Band ??? ms");
			} else {
				sprintf(buffer, "Doodah Band ~%.1f ms", average_milliseconds);
			}
			SDL_SetWindowTitle(window, buffer);

			frames_sum = 0;
			frame_milliseconds_sum = 0;
		}

		// Update screen.
		if (glGetError() != 0) {
			fprintf(stderr, "ERROR: OpenGL error\n");
			abort();
		}
		SDL_GL_SwapWindow(window);

		// Regulate frame rate.
		const uint64_t now = SDL_GetTicks64();
		if (now-frame_start < frame_ms) {
			SDL_Delay(frame_ms-(now-frame_start));
		}
		if (!paused) time_seconds += (float)(now-frame_start)/1000.0;
	}

	// Tidy everything.
	SDL_GL_DeleteContext(context);
	SDL_CloseAudioDevice(audio_device);
	SDL_DestroyWindow(window);
	SDL_Quit();
	// close(listen_socket);
	return 0;
}
